/*
This module parses HTML and anything that looks like it, i.e. is
composed of <tags> and text. That means it should be able to read
XML and similar as well. It doesn't interpret what it sees at all.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <Mowitz/MwUtils.h>
#include <Mowitz/MwFormat.h>
#include "MwHtmlParser.h"

static MwFmt fmt0 = {
        "Times",    /* font family */
        120,            /* size in decipoints */
        0,              /* bold */
        0,              /* italic */
        0,              /* underline */
        0,              /* strikethrough */
        "black",        /* foreground */
        "white",        /* background */
        0,              /* borders */
        MW_VADJ_CENTER, /* vertical adjust */
        MW_HADJ_LEFT,   /* horizontal adjust */
        0};             /* style */

static MwFmt fmt_h1 = {"Helvetica", 240, 1, 0, 0, 0, "black", "white",
                        0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h2 = {"Helvetica", 180, 1, 0, 0, 0, "black", "white",
                        0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h3 = {"Helvetica", 160, 1, 0, 0, 0, "black", "white",
                        0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h4 = {"Helvetica", 140, 1, 0, 0, 0, "black", "white",
                        0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h5 = {"Helvetica", 120, 1, 0, 0, 0, "black", "white",
                        0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};
static MwFmt fmt_h6 = {"Helvetica", 100, 1, 0, 0, 0, "black", "white",
                        0, MW_VADJ_CENTER, MW_HADJ_LEFT, 0};

enum {LIST_NONE=0, LIST_UL, LIST_OL, LIST_DL};

typedef struct hstate {
	int quiet;			/* don't output anything */
        int pre;                        /* preformatted, default 0 = false */
        int list;                       /* list mode, default LIST_NONE */
        int ll;                         /* list level, starts with 1 */
        int indent;                     /* indentation level, default 0 */
        int fmt;                        /* format */
        void *closure;
} hstate;

struct html_info {
	char *b;
	int n, nmax;
	hstate h;
	struct tagstack *ts;
	object_box *ob;
};

struct tagstack {
	char *tag;
	char *text;
	void (*atend)(struct html_info *);
	hstate h;
	struct tagstack *next;
};

/* convert string in place */
static void unquote(char *p)
{
	int state = 0;
	int c;
	char *q, *s;

	q = s = p;
	while (*p) {
		switch (state) {
		case 0:
			if (*p == '&') {
				state = 1;
				s = p+1;
			} else {
				*q++ = *p;
			}
			p++;
			break;
		case 1:
			if (*p == ';') {
				*p = '\0';
				c = MwFromCchar(s);
				if (c == -1) {	/* no match found */
					sprintf(q, "&%s;", s);
					q += strlen(q);
				} else {
					*q++ = c;
				}
				state = 0;
			}
			p++;
			break;
		}
	}
	*q = '\0';
}

void MwHtmlFree(object_box *ob)
{
	object_box *ob1, *ob2;

	for (ob1 = ob; ob1; ob1 = ob2) {
		ob2 = ob1->next;
		switch (ob1->type) {
		case MW_HTML_SPACE:
		case MW_HTML_WORDPART:
			MwFree(ob1->data);
			break;
		default:
			break;
		}
		MwFree(ob1);
	}
}


static MwHtmlNode *last_node(MwHtmlNode *h)
{
	if (h == NULL) return NULL;
	if (h->next == NULL) return h;
	return last_node(h->next);
}

static void grow_b(struct html_info *p, int n)
{
	if (n >= p->nmax) {
		p->nmax = n+1000;
		p->b = MwRealloc(p->b, p->nmax+1);
	}
}

static void clear_b(struct html_info *p)
{
	p->n = 0;
	p->b[0] = '\0';
}

static void emitchar(struct html_info *p, int c)
{
	grow_b(p, p->n+1);
	p->b[p->n++] = c;
	p->b[p->n] = '\0';
}

static void add_object_box(struct html_info *p, int type, void *data)
{
	object_box *ob = MwMalloc(sizeof *ob);
	ob->type = type;
	ob->data = data;
	ob->next = p->ob;
	p->ob = ob;
}

static void newline(struct html_info *p)
{
	add_object_box(p, MW_HTML_NEWLINE, NULL);
}

static void emitspace(struct html_info *p)
{
	char *q, *r;

	if (p->n == 0) return;
	if (p->h.quiet == 0) {
		if (p->h.pre) {
			q = p->b;
			while ((r = strchr(q, '\n'))) {
				*r = '\0';
				add_object_box(p, MW_HTML_WORDPART,
						MwRcMakerich(q, p->h.fmt));
				newline(p);
				q = r+1;
			}
			if (*q) add_object_box(p, MW_HTML_WORDPART,
						MwRcMakerich(q, p->h.fmt));
printf("Should be preformatted\n");
		} else {
			add_object_box(p, MW_HTML_SPACE, MwRcMakerich(p->b, p->h.fmt));
		}
	}
	clear_b(p);
}

static void tag_a(struct html_info *);
static void tag_address(struct html_info *);
static void tag_area(struct html_info *);
static void tag_b(struct html_info *);
static void tag_body(struct html_info *);
static void tag_br(struct html_info *);
static void tag_dd(struct html_info *);
static void tag_dl(struct html_info *);
static void tag_dt(struct html_info *);
static void tag_h1(struct html_info *);
static void tag_h2(struct html_info *);
static void tag_h3(struct html_info *);
static void tag_h4(struct html_info *);
static void tag_h5(struct html_info *);
static void tag_h6(struct html_info *);
static void tag_head(struct html_info *);
static void tag_hr(struct html_info *);
static void tag_html(struct html_info *);
static void tag_i(struct html_info *);
static void tag_img(struct html_info *);
static void tag_li(struct html_info *);
static void tag_map(struct html_info *);
static void tag_ol(struct html_info *);
static void tag_p(struct html_info *);
static void tag_pre(struct html_info *);
static void tag_table(struct html_info *);
static void tag_td(struct html_info *);
static void tag_th(struct html_info *);
static void tag_tr(struct html_info *);
static void tag_ul(struct html_info *);
static void tag_unknown(struct html_info *);

static struct {
	char *name;
	void (*action)(struct html_info *);
} tag[] = {
	{"A", tag_a},
	{"ADDRESS", tag_address},
        {"AREA", tag_area},
        {"B", tag_b},
        {"BR", tag_br},
        {"BODY", tag_body},
        {"DD", tag_dd},
        {"DL", tag_dl},
        {"DT", tag_dt},
        {"H1", tag_h1},
        {"H2", tag_h2},
        {"H3", tag_h3},
        {"H4", tag_h4},
        {"H5", tag_h5},
        {"H6", tag_h6},
        {"HEAD", tag_head},
        {"HR", tag_hr},
        {"HTML", tag_html},
        {"I", tag_i},
        {"IMG", tag_img},
        {"LI", tag_li},
        {"MAP", tag_map},
        {"OL", tag_ol},
        {"P", tag_p},
        {"PRE", tag_pre},
        {"TABLE", tag_table},
        {"TD", tag_td},
        {"TH", tag_th},
        {"TR", tag_tr},
        {"UL", tag_ul},
	{NULL, tag_unknown}
};

static void indent(struct html_info *p, int n)
{
	p->h.indent += n;
	add_object_box(p, MW_HTML_INDENT, (void *)(p->h.indent));
}

/* the indentation level is automatically reset after a list, but we
   need this to tell the widget about the new value */
static void unindent(struct html_info *p)
{
	indent(p, 0);
	newline(p);
}

static void end_anchor(struct html_info *p)
{
	add_object_box(p, MW_HTML_END_A, NULL);
}

/* Anchor tag */
static void tag_a(struct html_info *p)
{
	MwHtmlAnchor *a = malloc(sizeof *a);
	MwFmt fmt;
	MwDecodeFormat(p->h.fmt, ~0, &fmt);
	fmt.uline = 1;
	fmt.fg = "blue";
	p->h.fmt = MwEncodeFormat(~0, &fmt);
	a->href = MwHtmlGetValue("HREF", p->ts->text);
	a->name = MwHtmlGetValue("NAME", p->ts->text);
	add_object_box(p, MW_HTML_BEGIN_A, a);
	p->ts->atend = end_anchor;
	/* don't do the anchor right now */
}

/* Address tag. Only changes font */
static void tag_address(struct html_info *p)
{
	MwFmt fmt;
	MwDecodeFormat(p->h.fmt, ~0, &fmt);
	fmt.italic = 1;
	p->h.fmt = MwEncodeFormat(~0, &fmt);
}

/* Treat these as <a> */
static void tag_area(struct html_info *p)
{
	;
}

static void tag_b(struct html_info *p)
{
	MwFmt fmt;
	MwDecodeFormat(p->h.fmt, ~0, &fmt);
	fmt.bold = 1;
	p->h.fmt = MwEncodeFormat(~0, &fmt);
}

/* Check for background values et al */
static void tag_body(struct html_info *p)
{
	;
}

/* Line break */
static void tag_br(struct html_info *p)
{
	newline(p);
}

/* description in description list */
static void tag_dd(struct html_info *p)
{
	indent(p, 1);
	newline(p);
}

/* start description list */
static void tag_dl(struct html_info *p)
{
	p->h.list = LIST_DL;
	p->h.ll = 1;
	indent(p, 1);
	newline(p);
	p->ts->atend = unindent;
}

/* tag in description list */
static void tag_dt(struct html_info *p)
{
	indent(p, -1);
	newline(p);
}

/* all headers set font, start new line, print children and start new line */
/* Same problem again: how to get newline after header? */
static void tag_h1(struct html_info *p)
{
	newline(p);
	p->h.fmt = MwEncodeFormat(~0, &fmt_h1);
	p->ts->atend = newline;
}

static void tag_h2(struct html_info *p)
{
	newline(p);
	p->h.fmt = MwEncodeFormat(~0, &fmt_h2);
	p->ts->atend = newline;
}

static void tag_h3(struct html_info *p)
{
	newline(p);
	p->h.fmt = MwEncodeFormat(~0, &fmt_h3);
	p->ts->atend = newline;
}

static void tag_h4(struct html_info *p)
{
	newline(p);
	p->h.fmt = MwEncodeFormat(~0, &fmt_h4);
	p->ts->atend = newline;
}

static void tag_h5(struct html_info *p)
{
	newline(p);
	p->h.fmt = MwEncodeFormat(~0, &fmt_h5);
	p->ts->atend = newline;
}

static void tag_h6(struct html_info *p)
{
	newline(p);
	p->h.fmt = MwEncodeFormat(~0, &fmt_h6);
	p->ts->atend = newline;
}

/* ignore this tag and all children */
/* Set a flag that this is to be ignored */
static void tag_head(struct html_info *p)
{
printf("tag_head()\n");
	p->h.quiet = 1;
}

/* Horizontal ruler */
static void tag_hr(struct html_info *p)
{
	newline(p);
	add_object_box(p, MW_HTML_HR, NULL);
	newline(p);
}

/* this tag doesn't do anything per se */
static void tag_html(struct html_info *p)
{
	;
}

static void tag_i(struct html_info *p)
{
	MwFmt fmt;
	MwDecodeFormat(p->h.fmt, ~0, &fmt);
	fmt.italic = 1;
	p->h.fmt = MwEncodeFormat(~0, &fmt);
}

/* Image tag. Ignored until we have somewhere to show pictures. */
static void tag_img(struct html_info *p)
{
	char b[1024];
	MwHtmlImage *a = MwMalloc(sizeof *a);
	char *q = MwHtmlGetValue("ALT", p->ts->text);
	if (q == NULL) strcpy(b, "[INLINE]");
	else strcpy(b, q);
	a->src = MwHtmlGetValue("SRC", p->ts->text);
	a->alt = MwHtmlGetValue("ALT", p->ts->text);
	MwFree(q);
	add_object_box(p, MW_HTML_IMAGE, a);
}

/* start a new line. Should also fiddle with indentation */
/* Need to create bullets, but how? */
static void tag_li(struct html_info *p)
{
	newline(p);
	switch (p->h.list) {
	case LIST_UL:
	add_object_box(p, MW_HTML_UBULLET, (void *)'*');
		break;
	case LIST_OL:
	add_object_box(p, MW_HTML_OBULLET, (void *)p->h.ll);
		break;
	default:
	add_object_box(p, MW_HTML_UBULLET, (void *)' ');
		break;
	}
	p->h.ll++;
}

/* Not much to do here... */
static void tag_map(struct html_info *p)
{
	;
}

static void tag_ol(struct html_info *p)
{
	p->h.list = LIST_OL;
	p->h.ll = 1;
	indent(p, 1);
	newline(p);
	p->ts->atend = unindent;
}

/* Paragraph break */
static void tag_p(struct html_info *p)
{
	newline(p);
}

/* Set a flag that this is preformatted */
static void tag_pre(struct html_info *p)
{
	MwFmt fmt;
	MwDecodeFormat(p->h.fmt, ~0, &fmt);
	fmt.family = "Courier";
	p->h.fmt = MwEncodeFormat(~0, &fmt);
	p->h.pre = 1;
	newline(p);
	p->ts->atend = newline;
}

/* We don't really do tables just yet */
static void tag_table(struct html_info *p)
{
	;
}

/* Table cell */
static void tag_td(struct html_info *p)
{
	add_object_box(p, MW_HTML_TAB, NULL);
}

/* Table header */
static void tag_th(struct html_info *p)
{
	add_object_box(p, MW_HTML_TAB, NULL);
}

/* Table row */
static void tag_tr(struct html_info *p)
{
	newline(p);
}

static void tag_ul(struct html_info *p)
{
	p->h.list = LIST_UL;
	p->h.ll = 1;
	indent(p, 1);
	newline(p);
	p->ts->atend = unindent;
}

/* Handles tags we don't have handlers for. Ignore the tag. */
static void tag_unknown(struct html_info *p)
{
printf("tag_unknown() ignores tag '%s'\n", p->ts->tag);
}

static void emittag(struct html_info *p)
{
	int i;
	void (*action)(struct html_info *p);
	char *s, *t;
	struct tagstack *ts;
	if (p->n == 0) return;

	unquote(p->b);
	p->n = strlen(p->b);
	t = p->b+1;		/* skip leading < */
	p->b[p->n-1] = '\0';	/* strip trailing > */
	s = strchr(t, ' ');
	if (s) *s++ = '\0';	/* split tag and parameters */

	if (t[0] == '/') {	/* end tag, look for match in stack */
		t++;
		while (p->ts && MwStrcasecmp(p->ts->tag, t)) {
			ts = p->ts;
			p->ts = ts->next;
			MwFree(ts->tag);
			MwFree(ts->text);
		}
		if (p->ts == NULL) {
printf("Warning: bogus closer '%s'\n", t);
			clear_b(p);
			return;
		}
		ts = p->ts;
		p->ts = ts->next;
		p->h = ts->h;
		MwFree(ts->tag);
		MwFree(ts->text);
		action = ts->atend;
		MwFree(ts);
		if (action) (*action)(p);
if (!MwStrcasecmp(t, "head"))
printf("t=%s, pre=%d, quiet=%d\n", t, p->h.pre, p->h.quiet);

		clear_b(p);
		return;
	}

	/* we're here, so it's a start tag */
	ts = MwMalloc(sizeof(*ts));
	if (s == NULL) s = "";
	ts->tag = MwStrdup(t);
	ts->text = MwStrdup(s);
	ts->atend = NULL;
	ts->h = p->h;
	ts->next = p->ts;
	p->ts = ts;
	for (i = 0; tag[i].name; i++) {
		if (!MwStrcasecmp(tag[i].name, t)) break;
	}
	action = tag[i].action;

	(*action)(p);
	clear_b(p);
}

static void emitword(struct html_info *p)
{
	if (p->n == 0) return;
	if (p->h.quiet == 0) {
		unquote(p->b);
		add_object_box(p, MW_HTML_WORDPART, MwRcMakerich(p->b, p->h.fmt));
	}
	clear_b(p);
}

object_box *MwHtmlParse(int (*getc_cb)(void *), void *closure)
{
	enum {START = 0, SPACE, NEWLINE1, NEWLINE2, ENDTAG,
		BEGIN_TAG, STARTTAG, NEWLINE3, WORDPART, END} state;
	int error;
	int c;
	struct html_info p;
	object_box *ob, *ob1;

	p.n = p.nmax = 0;
	p.b = NULL;
	grow_b(&p, 1);
	p.ts = NULL;
	p.h.pre = 0;
	p.h.quiet = 0;
	p.h.indent = 0;
	p.h.list = LIST_NONE;
	p.h.ll = 0;
	p.h.fmt = MwEncodeFormat(~0, &fmt0);
	p.h.closure = closure;
	p.ob = NULL;

	state = START;
	error = 0;
	while (state != END) {
		c = getc_cb(closure);
/*printf("state = %d, c = '%c'\n", state, c);*/
		switch (state) {
		case START:
			if (c == '\0') {
				state = END;
			} else if (c == '\n') {
				/* must ignore newline before end tag */
				state = NEWLINE1;
			} else if (c == '<') {
				emitchar(&p, c);
				state = BEGIN_TAG;
			} else if (isspace(c)) {
				emitchar(&p, c);
				state = SPACE;
			} else {
				emitchar(&p, c);
				state = WORDPART;
			}
			break;
		case SPACE:
			if (c == '\0') {
				emitspace(&p);
				state = END;
			} else if (c == '\n') {
				state = NEWLINE1;
			} else if (c == '<') {
				emitspace(&p);
				emitchar(&p, c);
				state = BEGIN_TAG;
			} else if (isspace(c)) {
				emitchar(&p, c);
			} else {
				emitspace(&p);
				emitchar(&p, c);
				state = WORDPART;
			}
			break;
		case NEWLINE1:
			if (c == '\0') {
				emitspace(&p);
				state = END;
			} else if (c == '<') {
				state = NEWLINE2;
			} else if (c == '\n') {
				emitchar(&p, '\n');
			} else if (isspace(c)) {
				emitchar(&p, '\n');
				emitchar(&p, c);
				state = SPACE;
			} else {
				emitchar(&p, '\n');
				emitspace(&p);
				emitchar(&p, c);
				state = WORDPART;
			}
			break;
		case NEWLINE2:
			if (c == '\0') {
				emitchar(&p, '\n');
				emitspace(&p);
				emitchar(&p, '<');
				emitword(&p);
				state = END;
			} else if (c == '/') {
				emitchar(&p, '\n');
				emitspace(&p);
				emitchar(&p, '<');
				emitchar(&p, c);
				state = ENDTAG;
			} else {
				emitspace(&p);
				emitchar(&p, '<');
				emitchar(&p, c);
				state = STARTTAG;
			}
			break;
		case ENDTAG:
			if (c == '\0') {
				emitword(&p);
				state = END;
			} else if (c == '>') {
				emitchar(&p, c);
				emittag(&p);
				state = START;
			} else {
				emitchar(&p, c);
			}
			break;
		case BEGIN_TAG:
			if (c == '\0') {
				emitspace(&p);
				state = END;
			} else if (c == '/') {
				emitchar(&p, c);
				state = ENDTAG;
			} else {
				emitchar(&p, c);
				state = STARTTAG;
			}
			break;
		case STARTTAG:
			if (c == '\0') {
				emitword(&p);
				state = END;
			} else if (c == '>') {
				emitchar(&p, c);
				emittag(&p);
				state = NEWLINE3;
			} else {
				emitchar(&p, c);
			}
			break;
		case NEWLINE3:
			if (c == '\0') {
                                state = END;
                        } else if (c == '\n') {
				state = START;
			} else if (c == '<') {
				emitchar(&p, c);
				state = BEGIN_TAG;
			} else if (isspace(c)) {
				emitchar(&p, c);
				state = SPACE;
			} else {
				emitchar(&p, c);
				state = WORDPART;
			}
			break;
		case WORDPART:
			if (c == '\0') {
				emitword(&p);
                                state = END;
                        } else if (c == '\n') {
				emitword(&p);
				state = NEWLINE1;
			} else if (isspace(c)) {
				emitword(&p);
				emitchar(&p, c);
				state = SPACE;
			} else if (c == '<') {
				emitword(&p);
				emitchar(&p, c);
				state = BEGIN_TAG;
			} else {
				emitchar(&p, c);
			}
			break;
		default:
			printf("foo\n");
			exit(0);
		}
	}
	add_object_box(&p, MW_HTML_NEWLINE, NULL);

/* reverse the object list */
ob = NULL;
while (p.ob) {
ob1 = p.ob;
p.ob = p.ob->next;
ob1->next = ob;
ob = ob1;
}

	if (error) {
		fprintf(stderr, "Error\n");
	}
	return ob;
}

/* p looks like this: key=value key2="value 2". Keys are not case
   sensitive. Everything not in quotes is converted to upper case.
   Returns value if found, otherwise NULL.
*/
char *MwHtmlGetValue(char *k, char *p)
{           
        char *q = p;
        int n = strlen(k);
        int i;
        char b[1024];
            
Again:
        while (*q && isspace(*q)) q++;  /* skip leading space */
        if (!MwStrncasecmp(k, q, n) && q[n] == '=') {     /* gotcha */
                n++;                                                
                i = 0;
                if (q[n] == '"') {
                        n++;      
                        while (q[n] && q[n] != '"') b[i++] = q[n++];
                } else {
                        while (q[n] && !isspace(q[n])) b[i++] = q[n++];
                }
                b[i] = '\0';
                return MwStrdup(b);
        }
        while (*q && !isspace(*q)) q++;
        if (*q) goto Again;
        return NULL;
}

